##############################################################################
#   Copyright (c) 2018 Eynes/E-MIPS (www.eynes.com.ar)
#   License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).
##############################################################################

from odoo import _, api, fields, models
from odoo.exceptions import ValidationError, UserError
from suds.client import Client
from odoo.addons.wsmail_ypf.utils.mail_wrapper import MIMEMailWrapper
from odoo.addons.wsmail_ypf.utils.mail_wrapper import DummyWrapper
import base64

class IrMailServer(models.Model):
    _name = "ir.mail_server"
    _inherit = "ir.mail_server"

    server_type = fields.Selection([
        ('ws', 'Web Service'),
        ('smtp', 'SMTP'),
    ], required=True, default='smtp')

    wstoken = fields.Char(
        string="Token",
    )

    @api.model
    def get_ws_methods(self, client):
        method_vals = {}

        for service in client.wsdl.services:
            for port in service.ports:
                methods = port.methods.values()
                for method in methods:
                    method_name = str(method.name)
                    method_vals[method_name] = []
                    for part in method.soap.input.body.parts:
                        part_type = part.type
                        if(not part_type):
                            part_type = part.element[0]
                        o = client.factory.create(part_type)
                        params = [str(i) for i in o.__dict__.keys() if i[:1] != '_']
                        method_vals[method_name].extend(params)
        return method_vals

    @api.model
    def get_ws_method(self, methods, params):
        """Get Ws Method by parameters

        @param param:  Parameters to match
        @type  param:  List

        @return:  WebService Method Name
        @rtype :  String

        """

        web_service = None
        for key, mt_params in methods.items():
            if 'secHash' not in mt_params:
                if 'secHash' in params:
                    params.remove('secHash')
            sparam = set(mt_params)
            sparams = set(params)
            if len(sparam) != len(sparams):
                continue
            add_it = sparams - sparam
            rem_it = sparam - sparams
            if not add_it and not rem_it:
                web_service = key
                break
        print(web_service, params)
        return web_service

    @api.multi
    def test_smtp_connection(self):
        ws_server = self.filtered(
                lambda x: x.server_type == 'ws')
        smtp_server = self.filtered(
                lambda x: x.server_type == 'smtp')
        if smtp_server and ws_server:
            raise ValidationError(
                _("You cannot test connection of "\
                  "SMTP and Web Service at the same time"))
        if smtp_server:
            return super(IrMailServer, self).test_smtp_connection()

        for server in self:
            try:
                client = self.connect(
                    server.smtp_host,
                    mail_server_id=server.id,
                )
            except Exception as e:
                raise UserError(
                    _("Connection Test Failed!"))
            if not client.wsdl.services:
                raise UserError(
                    _("Connection Test Failed! No services found"))
        raise UserError(
                _("Connection Test Succeeded! Everything seems properly set up!"))

    def connect(self, host=None, port=None, user=None, password=None, encryption=None,
                smtp_debug=False, mail_server_id=None):

        mail_server = self.sudo().browse(mail_server_id)
        if not mail_server and not host:
            mail_server = self.sudo().search([], order='sequence', limit=1)

        smtp_pass = password or mail_server.smtp_pass
        dummy_ctx = {
            'smtp_pass': smtp_pass
        }

        if mail_server:
            if mail_server.server_type != 'ws':
                return super(IrMailServer, self).connect(
                    host=host, port=port, user=user, password=password,
                    encryption=encryption, smtp_debug=smtp_debug,
                    mail_server_id=mail_server_id)
            else:
                token = mail_server.wstoken
                client = Client(mail_server.smtp_host)
                client.set_options(
                    headers={
                        'Ocp-Apim-Subscription-Key': token,
                    }
                )
                return DummyWrapper(client, ctx=dummy_ctx)

        if '?wsdl' not in host.lower():
            return super(IrMailServer, self).connect(
                host=host, port=port, user=user, password=password,
                encryption=encryption, smtp_debug=smtp_debug,
                mail_server_id=mail_server_id)
        else:
            #Necesitamos el mail_server para sacar el token
            if not mail_server:
                mail_server = self.sudo().search([
                    ('smtp_host', '=', host),
                ], limit=1)
            token = mail_server.wstoken
            client = Client(host)
            client.set_options(
                headers={
                    'Ocp-Apim-Subscription-Key': token,
                }
            )
            return DummyWrapper(client, ctx=dummy_ctx)


    def build_email(self, email_from, email_to, subject, body, email_cc=None,
                    email_bcc=None, reply_to=False, attachments=None,
                    message_id=None, references=None, object_id=False,
                    subtype='plain', headers=None, body_alternative=None,
                    subtype_alternative='plain'):
        msg = super(IrMailServer, self).build_email(
                email_from, email_to, subject, body, email_cc=email_cc,
                email_bcc=email_bcc, reply_to=reply_to, attachments=attachments,
                message_id=message_id, references=references, object_id=object_id,
                subtype=subtype, headers=headers, body_alternative=body_alternative,
                subtype_alternative=subtype_alternative)

        message_vals = {
            'to': email_to,
            'from': email_from,
            'subject': subject,
        }

        if email_cc:
            message_vals.update({
                'cc': email_cc
            })

        if email_bcc:
            message_vals.update({
                'bcc': email_bcc
            })

        if subtype == 'html':
            message_vals.update({
                'bodyContent': body,
                'bodySignature':  True,
            })
        else:
            message_vals.update({
                'body': body
            })

        buffer_attach = []
        buffer_name = []
        if attachments:
            for (fname, fcontent, mime) in attachments:
                data = base64.b64encode(fcontent)
                buffer_name.append(fname)
                buffer_attach.append(data)

            message_vals.update({
                'bufferAttachments': buffer_attach,
                'nameAttachments': buffer_name,
            })

        return MIMEMailWrapper(message_vals, msg)

    def send_ws_mail(self, session, method, message):
        service = getattr(session.service, method, False)
        if not service:
            raise ValidationError(_("Invalid Method"))
        try:
            vals = message.get_msg_vals()
            res = service(**vals)
        except Exception as e:
            raise ValidationError(e)
        return message['Message-Id']

    @api.model
    def send_email(self, message, mail_server_id=None, smtp_server=None, smtp_port=None,
                   smtp_user=None, smtp_password=None, smtp_encryption=None,
                   smtp_debug=False, smtp_session=None):
        mail_server = self.sudo().browse(mail_server_id)

        session = smtp_session
        if isinstance(session, DummyWrapper) or \
                mail_server.server_type == 'ws' or\
                '?wsdl' in (smtp_server or '').lower():

            if not session:
                if mail_server.server_type == 'ws':
                    session = self.connect(mail_server_id=mail_server.id)
                else:
                    session = self.connect(host=smtp_server)

            if not session.wsdl.services:
                raise ValidationError(_("No services found"))

            if mail_server:
                sec_hash = mail_server.smtp_pass
            elif smtp_password:
                sec_hash = smtp_password
            else:
                sec_hash = session.ctx.get('smtp_pass')

            methods = self.get_ws_methods(session)
            if sec_hash:
                message.secHash = sec_hash
            values = message.get_msg_vals()
            ws_params = list(values.keys())
            method = self.get_ws_method(methods, ws_params)
            return self.send_ws_mail(session, method, message)

        return super(IrMailServer, self).send_email(
                message, mail_server_id=mail_server_id, smtp_server=smtp_server,
                smtp_port=smtp_port, smtp_user=smtp_user, smtp_password=smtp_password,
                smtp_encryption=smtp_encryption, smtp_debug=smtp_debug,
                smtp_session=smtp_session)


    @api.onchange('server_type')
    def onchange_server_type(self):
        if self.server_type == 'ws':
            self.smtp_encryption = 'none'
